#define MODULE lists
/* Handle List Operations where such operations are nontrivial */

/* Convert an index returned by the list selection box to a Menu Index
 * This is necessary because there may be holes in the MenuList.
 */

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "lists.h"


//*******************************************************
//*							*
//*	ListBase function definitions			*
//*							*
//*******************************************************

BOOL
ListBase::HasElement(void const*pvData) const
{
	ListData* const*	ppld;

	for (ppld = &pld ; *ppld && (*ppld)->pvData != pvData; ppld = &(*ppld)->pldNext)
		;

	return 0 != *ppld;
}

void
ListBase::AddElement(void *pvData)
{
	ListData *pldNew = new ListData(pvData, 0);

	if (pldTail)
	{
		pldTail->pldNext = pldNew;
		pldTail = pldTail->pldNext;
	}
	else
		pld = pldTail = pldNew;
}

void
ListBase::PushElement(void *pvData)
{
	ListData *pldNew = new ListData(pvData, 0);

	if (pld)
	{
		pldNew->pldNext = pld;
		pld = pldNew;
	}
	else
		pld = pldTail = pldNew;
}

void
ListBase::InsertElement(int iLocation, void *pvData)
{
	int		i;
	ListData	**ppld;

	if (iLocation <= 0)
		PushElement(pvData);
	else
	{
		for (ppld = &pld, i = 0;
		     *ppld && i < iLocation - 1;
		     ppld = &(*ppld)->pldNext)
		{
			i++;
		}
		if (*ppld)
		{
			ListData *pldNew = new ListData(pvData, 0);
			pldNew->pldNext = (*ppld)->pldNext;
			(*ppld)->pldNext = pldNew;
			if (*ppld == pldTail)
				pldTail = pldNew;
		}
		else
			AddElement(pvData);
	}
}

BOOL
ListBase::RemoveElement(void *pvData)
{
	ListData	**ppld, **ppldLast, *pldRemove;

	for (ppld = &pld, ppldLast = 0;
	     *ppld && (*ppld)->pvData != pvData;
	     ppld = &(*ppld)->pldNext)
		ppldLast = ppld;
	if (0 == *ppld)
		// not found - nothing removed
		return FALSE;

	pldRemove = *ppld;
	*ppld = pldRemove->pldNext;
	if (pldRemove == pldTail)
		pldTail = (ppldLast ? *ppldLast : 0);
	delete pldRemove;

	return TRUE;
}

void
ListBase::RemoveIndex(int iIndex)
{
	ListData	**ppld, **ppldLast, *pldRemove;
	int		i;

	for (ppld = &pld, ppldLast = 0, i=0;
	     *ppld && i != iIndex;
	     ppld = &(*ppld)->pldNext,i++)
		ppldLast = ppld;
	if (*ppld)
	{
		pldRemove = *ppld;
		*ppld = pldRemove->pldNext;
		if (pldRemove == pldTail)
			pldTail = (ppldLast ? *ppldLast : 0);
		delete pldRemove;
	}
}

void
ListBase::SortElements(int (* comparison_function)(const void * v0,const void *v1))
{
	int		list_size;
	ListData	*next_list_data;
	void		**item_list;
	int		count;

	/* Allocate sorting memory for the operation */

	list_size = Size();
	item_list = new void * [list_size];

	/* Generate a list of ListData pointers to sort */

	count = 0;
	next_list_data = pld;
	while (next_list_data != NULL)
	{
		item_list[count++] = next_list_data->pvData;
		next_list_data = next_list_data->pldNext;
	};

	/* And now sort the array of pointers to whatever */

	qsort((void *)item_list,list_size,sizeof(void *),comparison_function);

	/* And now place the sorted data back into the list data structure */

	count = 0;
	next_list_data = pld;
	while (next_list_data != NULL)
	{
		next_list_data->pvData = item_list[count++];
		next_list_data = next_list_data->pldNext;
	};

	/* Free any allocated memory */

	delete [] item_list;

}

void *
ListBase::PopHead(void)
{
	void	*pvData;
	ListData *pldOld;

	if (!pld)
		return 0;
	pldOld = pld;
	pld = pld->pldNext;
	if (!pld)
		pldTail = 0;
	pvData = pldOld->pvData;
	delete pldOld;
	return pvData;
}

void
ListBase::Clear(void)
{
	while( PopHead() )
		;
}

int
ListBase::Size(void) const
{
	ListData	* const *ppld;
	int	nSize = 0;

	for (ppld = &pld;
	     *ppld;
	     ppld = &(*ppld)->pldNext, nSize++);
	return nSize;
}

void
ListBase::LocationExchange(int iLocation)
{
	ListData	**ppld, **ppldNext;
	void		*pvData;
	int		i;

	for (ppld = &pld, i = 0; *ppld && i != iLocation; ppld = &(*ppld)->pldNext, i++);
	if (i == iLocation && *ppld && (*ppld)->pldNext)
	{
		ppldNext = &(*ppld)->pldNext;
		pvData = (*ppld)->pvData;
		(*ppld)->pvData = (*ppldNext)->pvData;
		(*ppldNext)->pvData = pvData;
	}
}
