#ifndef __LISTS_H
#define __LISTS_H

#ifndef __ASSERT_H
#include <assert.h>
#endif

#ifndef __COMPARE_H
#include "compare.h"
#endif

#ifndef __STRARRAY_H
#	include "strarray.h"
#endif

// force a UAE
#ifndef CRASH
#	define CRASH	*((char *) 0) = '\0'
#endif

void	BuildMenuList(STRING_ARRAY &menuList);
int	ListToMenu(int ListIndex);
void	CopyList(char **&Dest, char **Source);


//***********************************************************************
//*									*
//*	ListDataUser is a wrapper for the protected class, ListData	*
//*									*
//***********************************************************************

class	ListDataUser
{
protected:

	struct	ListData
	{
		void	*pvData;
		ListData *pldNext;

		inline	ListData(	void *pv,
					ListData *pld) :
				pvData(pv),
				pldNext(pld)
		{
		}

	};

	ListData	*pld;
	ListData	*pldTail;

	inline	ListDataUser(void) : pld(0), pldTail(0) {}
	inline	ListDataUser(const ListDataUser& that) : pld(that.pld), pldTail(that.pldTail) {}
	inline	ListDataUser(ListData *pldInitial) :
		pld(pldInitial),
		pldTail(pldInitial)
		{}
};


//***********************************************************************
//*									*
//*	ListBase is the root of all our List classes			*
//*									*
//***********************************************************************

class	ListBase : public ListDataUser
{
private:
	friend	class WalkerBase;
	friend	class ConstWalkerBase;

protected:
	inline	ListBase(void) : ListDataUser() {}

	BOOL	HasElement(void const*pvData) const;
	void	AddElement(void *pvData);
	void	PushElement(void *pvData);
	void	InsertElement(int iLocation, void *pvData);
	BOOL	RemoveElement(void *pvData);
	void	RemoveIndex(int iIndex);
	void	SortElements(int (* comparison_function)(const void * v0,const void *v1));
	void	LocationExchange(int iLocation);

	inline	void	*GetHead(void) { return pld ? pld->pvData : 0; }
	inline	const void	*GetHead(void) const { return pld ? pld->pvData : 0; }
	void	*PopHead(void);

public:
	int	Size(void) const;
	void	Clear(void);
};

//***********************************************************************
//*									*
//*	Walker is the base for out List walker classes			*
//*									*
//***********************************************************************

class	WalkerBase : public ListDataUser
{
protected:
	inline	WalkerBase(void) : ListDataUser() {}
	inline	WalkerBase(const ListBase &lb) : ListDataUser(lb.pld) {}
	inline	WalkerBase(const ListBase *lb) : ListDataUser(lb->pld) {}

	inline	void		*GetHead(void) { return pld ? pld->pvData : 0; }
	inline	void const	*GetHead(void) const { return pld ? pld->pvData : 0; }
};

//***********************************************************************
//*									*
//*	Walker is the base for out List walker classes			*
//*									*
//***********************************************************************

class	ConstWalkerBase : public ListDataUser
{
protected:
	inline	ConstWalkerBase(void) : ListDataUser() {}
	inline	ConstWalkerBase(const WalkerBase& w) : ListDataUser(w) {}
	inline	ConstWalkerBase(const ListBase &lb) : ListDataUser(lb.pld) {}
	inline	ConstWalkerBase(const ListBase *lb) : ListDataUser(lb->pld) {}

	inline	const void	*GetHead(void) const { return pld ? pld->pvData : 0; }
};

template <class T> class Walker;
template <class T> class ConstWalker;

//***********************************************************************
//*									*
//*	The templates to make it all type safe				*
//*									*
//***********************************************************************

template<class T>
class Walker;

template<class T>
class ConstWalker;

template <class T> class List :
//public Equivalent <List <T> >,
public ListBase
{
	friend class Walker<T>;
	friend class ConstWalker<T>;
public:
	typedef Walker<T> Walker_t;
	typedef ConstWalker<T> ConstWalker_t;

public:
	typedef T Element;

	List(int = 0) {}

	int operator == (const List &l) const
	{
		return (GetHead() == l.GetHead());
	}

	inline	BOOL
	HasElement(T const*pData) const
	{
		return ListBase::HasElement((void const*) pData);
	}

	inline	void
	AddElement(T *pData)
	{
		ListBase::AddElement((void *) pData);
	}

	inline	void
	PushElement(T *pData)
	{
		ListBase::PushElement((void *) pData);
	}

	inline	void
	InsertElement(int iLocation, T *pData)
	{
		ListBase::InsertElement(iLocation, (void *) pData);
	}

	inline	BOOL
	RemoveElement(T *pData)
	{
		return ListBase::RemoveElement((void *) pData);
	}

	inline	T	*GetHead(void) { return (T *) ListBase::GetHead(); }
	inline	T const *GetHead(void) const { return (T *) ListBase::GetHead(); }
	inline	T	*PopHead(void) { return (T *) ListBase::PopHead(); }

	inline	void
	SortElements(int (* comparison_function)(const void * v0,const void *v1))
	{
		ListBase::SortElements(comparison_function);
	}

	inline	void
	LocationExchange(int iLocation)
	{
		ListBase::LocationExchange(iLocation);
	}
};

template<class T>
void
RemoveAllElements(List<T>& l)
{
	T* p;
	while ( p = l.PopHead() )
		delete p;
}

template<class T> class SimpleWalker;
template<class T> class ConstSimpleWalker;

//***********************************************************************
//*									*
//*	The templates to make it all type safe				*
//*									*
//***********************************************************************

template <class T>
class SimpleWalker;

template <class T>
class ConstSimpleWalker;

template <class T> class SimpleList :
//public Equivalent <SimpleList<T> >,
public ListBase
{
	friend class SimpleWalker<T>;
	friend class ConstSimpleWalker<T>;
public:
	typedef SimpleWalker<T> Walker_t;
	typedef ConstSimpleWalker<T> ConstWalker_t;

public:
	typedef T Element;

	SimpleList(int = 0) {}

	int operator == (const SimpleList &l) const
	{
		return (GetHead() == l.GetHead());
	}

	inline	BOOL
	HasElement(T const*pData) const
	{
		return ListBase::HasElement((void const*) pData);
	}

	inline	void
	AddElement(T *pData)
	{
		ListBase::AddElement((void *) pData);
	}

	inline	void
	PushElement(T *pData)
	{
		ListBase::PushElement((void *) pData);
	}

	inline	void
	InsertElement(int iLocation, T *pData)
	{
		ListBase::InsertElement(iLocation, (void *) pData);
	}

	inline	BOOL
	RemoveElement(T *pData)
	{
		return ListBase::RemoveElement((void *) pData);
	}

	inline	T	*GetHead(void) { return (T *) ListBase::GetHead(); }
	inline	T const *GetHead(void) const { return (T *) ListBase::GetHead(); }
	inline	T	*PopHead(void) { return (T *) ListBase::PopHead(); }
};

template <class T> class Walker : public WalkerBase
{
public:
	/* Although these should have the inline keyword, BC++ complains if it
	 * is given.
	 */

	Walker(void) : WalkerBase() {}
	Walker(List<T> *lst) : WalkerBase(lst) {}
	Walker(List<T> &lst) : WalkerBase(lst) {}

	inline	Walker& operator=(const List<T> *lst) { pld = lst->pld; return *this; }
	inline	Walker& operator=(List<T> const &lst) { pld = lst.pld; return *this; }

	inline	bool operator ==(const Walker<T>& that) const { return GetHead() == that.GetHead(); }
	inline	operator T*(void) { return (T *) GetHead(); }

	inline	T &operator *(void) { return *(T *) GetHead(); }

	inline	T*	operator->(void) { return (T *) GetHead(); }

	inline	Walker& operator++(void)
	{
		pld = pld->pldNext;
		return *this;
	}

	inline	Walker	operator++(int)
	{
		Walker wTemp = *this;

		++(*this);

		return wTemp;
	}
};

template <class T> class ConstWalker : public ConstWalkerBase
{
public:
	ConstWalker(void) : ConstWalkerBase() {}
	ConstWalker(const Walker<T>& w) : ConstWalkerBase(w) {}
	ConstWalker(const List<T> *lst) : ConstWalkerBase(lst) {}
	ConstWalker(const List<T> &lst) : ConstWalkerBase(lst) {}

	inline	ConstWalker&	operator=(const List<T> &lst) { pld = lst.pld; return *this; }

	inline	operator const T*(void) const { return (const T *) GetHead(); }

	inline	const T &operator *(void) const { return *(const T *) GetHead(); }

	inline	const T*	operator->(void) const { return (const T *) GetHead(); }

	inline	ConstWalker&	operator++(void)
	{
		pld = pld->pldNext;
		return *this;
	}

	inline	ConstWalker	operator++(int)
	{
		ConstWalker wTemp = *this;

		++(*this);

		return wTemp;
	}
};

template <class T> class SimpleWalker : public WalkerBase
{
public:
	/* Although these should have the inline keyword, BC++ complains if it
	 * is given.
	 */

	SimpleWalker(void) : WalkerBase() {}
	SimpleWalker(SimpleList<T> *lst) : WalkerBase(lst) {}
	SimpleWalker(SimpleList<T> &lst) : WalkerBase(lst) {}

	inline	SimpleWalker&	operator=(const SimpleList<T> *lst) { pld = lst->pld; return *this; }
	inline	SimpleWalker&	operator=(const SimpleList<T> &lst) { pld = lst.pld; return *this; }

	inline	operator T*(void) { return (T *) GetHead(); }

	inline	T &operator *(void) { return *(T *) GetHead(); }

	inline	T*	operator++(void)
	{
		pld = pld->pldNext;
		return (T*) GetHead();
	}

	inline	T*	operator++(int)
	{
		T*	tTemp;

		tTemp = (T*) GetHead();
		pld = pld->pldNext;
		return tTemp;
	}
};

template <class T> class ConstSimpleWalker : public ConstWalkerBase
{
public:
	ConstSimpleWalker(void) : ConstWalkerBase() {}
	ConstSimpleWalker(const SimpleWalker<T>& w) : ConstWalkerBase(w) {}
	ConstSimpleWalker(const SimpleList<T> *lst) : ConstWalkerBase(lst) {}
	ConstSimpleWalker(const SimpleList<T> &lst) : ConstWalkerBase(lst) {}

	inline	ConstSimpleWalker&	operator=(const SimpleList<T> &lst) { pld = lst.pld; return *this; }

	inline	operator const T*(void) const { return (const T *) GetHead(); }

	inline	const T &operator *(void) const { return *(const T *) GetHead(); }

	inline	const T*	operator++(void)
	{
		pld = pld->pldNext;
		return (const T*) GetHead();
	}

	inline	const T*	operator++(int)
	{
		const T*	tTemp;

		tTemp = (const T*) GetHead();
		pld = pld->pldNext;
		return tTemp;
	}
};

/*---------------------------------------------------------------*/

template<class T>
class OwningList : public List<T>
{
	typedef OwningList<T>  this_t;
public:
	typedef ConstWalker<T> ConstWalker_t;
	typedef	     Walker<T>	    Walker_t;

	OwningList() {}
	OwningList( const this_t& rhs ) { operator=( rhs ); }
	OwningList& operator=( const this_t& rhs );
	~OwningList() { Clear(); }
	void Clear();
};

#ifndef __CNTFCT_H
#include "cntfct.h"
#endif

template<class T>
OwningList<T>& OwningList<T>::operator=( const OwningList<T>& rhs )
{
	Clear();
	::Append(*this,rhs);
	return *this;
}

template<class T>
void OwningList<T>::Clear()
{
	T* p;

	while( p = PopHead() )
		delete p;
}

#ifndef _MSC_VER
/*---------------------------------------------------------------*/
// ValList contains list values and not pointers to values as List does.
// I only implemented the functionality I needed. RAI 24-12-99

template<class T>
class ValList //: public Equivalent <ValList <T> >
{
public:
	typedef ValList<T> this_t;
	typedef T element_t;
	class ConstIterator;
	class Iterator;

private:
	class Node;

	Node* pnFirst;
	Node* pnLast;

public:
	ValList();
	ValList(const ValList<T>& );
	ValList& operator=(const ValList<T>& rhs);

	~ValList();

	void Clear();

	Iterator Begin () { return Iterator(pnFirst); }
	Iterator RBegin() { return Iterator(pnLast ); }
	Iterator End   () { return Iterator(0);	  }
	Iterator REnd  () { return Iterator(0);	  }

	ConstIterator Begin () const { return ConstIterator(pnFirst); }
	ConstIterator RBegin() const { return ConstIterator(pnLast ); }
	ConstIterator End   () const { return ConstIterator(0);	      }
	ConstIterator REnd  () const { return ConstIterator(0);	      }

	this_t& operator<<( const T& data );
	Iterator Add(const T& data ); // adds a new node at the end of the list
	void Append(const ValList<T>& rhs); // append another list at the end of the list
	Iterator Insert(Iterator w, const T& data);
	BOOL Erase(const T& val);
	Iterator Erase(Iterator w);
	Iterator Find( const T& data );
	BOOL IsExist( const T& data ) const;
	ConstIterator Find( const T& data ) const;

	BOOL Contains( const T& data ) const;

#ifdef USE_UNTESTED_CODE // don't use untested code!
	ConstIterator operator[](uint32 i) const
	{
		return (*const_cast<ValList<T>*>(this))[i];
	}

	Iterator operator[](uint32 i);
#endif

	BOOL Empty() const { return !pnFirst; }
	BOOL operator==( const this_t& rhs ) const;
	uint32 Size() const;

	this_t& Swap( this_t& rhs );
};

template<class T>
class ValList<T>::Node
{
	friend class ValList<T>;
	friend class ConstIterator;
private:
	ValList<T>::Node* pnPrev;
	ValList<T>::Node* pnNext;
	T data;

public:
	Node( const T& data_, Node* pnPrev_, Node* pnNext_);
	~Node();

	      T& Data()		{ return data; }
	const T& Data() const	{ return data; }
};

template<class T>
class ValList<T>::ConstIterator
{
	friend class ValList<T>;

	Node* pnBegin;

protected:
	T& Data();

public:
	ConstIterator( Node* pnBegin_ = 0 );
	ConstIterator& operator=( const ConstIterator& rhs );

	BOOL operator==(const ConstIterator& rhs) const;
	BOOL operator!=(const ConstIterator& rhs) const;

	const T& operator*() const;
	const T* operator->() const { return &operator*(); }

	ConstIterator& operator++();
	ConstIterator& operator--();
};

template<class T>
class ValList<T>::Iterator : public ValList<T>::ConstIterator
{
	typedef ConstIterator base_t;

public:
	Iterator( Node* pnBegin_ = 0 );
	Iterator& operator=( const Iterator& rhs );

	BOOL operator==(const Iterator& rhs) const;

	T& operator*();
	T* operator->() const { return &operator*(); }

	Iterator& operator++();
	Iterator& operator--();
};

template<class T>
ValList<T>& ValList<T>::Swap( ValList<T>& rhs )
{
	Node* p = pnFirst;
	pnFirst = rhs.pnFirst;
	rhs.pnFirst = p;

	p = pnLast;
	pnLast = rhs.pnLast;
	rhs.pnLast = p;

	return *this;
}

template<class T>
typename ValList<T>::ConstIterator ValList<T>::Find( const T& data ) const
{
	Iterator w = const_cast<ValList<T>* >(this)->Find(data);
	return w;
}

template<class T>
typename ValList<T>::Iterator ValList<T>::Erase(typename ValList<T>::Iterator w)
{
	Node* pnRet;

	if( w == End() )	// ignore Erase( End() )
		return w;

	if( !pnFirst )
	{
		assert( FALSE );	// signal in debug mode:
					//   wanted to Erase something != End() from an empty list
					// ignore in release mode
		return w;
	}

	if ( pnFirst == w.pnBegin )
	{
		pnFirst = w.pnBegin->pnNext;
	}

	if ( pnLast == w.pnBegin )
	{
		pnLast = w.pnBegin->pnPrev;
	}

	pnRet = w.pnBegin->pnNext;
	delete w.pnBegin;
	return pnRet;
}

template<class T>
BOOL ValList<T>::Erase(const T& val)
{
	Iterator w = Find( val );
	if ( w == End() )
		return FALSE;

	Erase( w );
	return TRUE;
}

template<class T>
typename ValList<T>::Iterator ValList<T>::Insert(typename ValList<T>::Iterator w, const T& data)
{
	if( w.pnBegin )
	{
		new Node( data, w.pnBegin->pnPrev, w.pnBegin );

		if( w.pnBegin == pnFirst )
			pnFirst = w.pnBegin->pnPrev;

		return --w;
	}
	// else
	return Add( data );
}

template<class T>
typename ValList<T>::Iterator ValList<T>::Add(const T& data ) // adds a new node at the end of the list
{
	pnLast = new Node( data, pnLast, 0 );
	if( !pnFirst )
		pnFirst = pnLast;

	return RBegin();
}

template<class T>
void ValList<T>::Append(const ValList<T>& rhs)
{
	for (ConstIterator w=rhs.Begin() ; w!=rhs.End() ; ++w)
	{
		Add (*w);
	}
}

template<class T>
uint32 ValList<T>::Size() const
{
	ConstIterator w;
	uint32 size = 0;

	for( w = Begin(); w != End(); ++w )
		++size;

	return size;
}

template<class T>
BOOL ValList<T>::operator==( const this_t& rhs ) const
{
	ConstIterator w, rhs_w;

	for( w = Begin(), rhs_w = rhs.Begin();
	     w != End() && rhs_w != rhs.End();
	     ++w, ++rhs_w )
		if( !(*w == *rhs_w) )
			return FALSE;

	return w == End() && rhs_w == rhs.End();
}

#ifdef USE_UNTESTED_CODE // don't use untested code!
template<class T>
typename ValList<T>::Iterator ValList<T>::operator[](uint32 i)
{
	Iterator w = Begin();
	while( i-- )
		++w;

	return w;
}
#endif

template<class T>
typename ValList<T>::Iterator ValList<T>::Find( const T& data )
{
	const Node* pn;

	for ( pn = pnFirst; pn; pn = pn->pnNext )
		if ( pn->data == data )
			return Iterator( const_cast<Node*>(pn) );

	return End();
}

template<typename T>
BOOL ValList<T>::IsExist( const T& data ) const

{
	const Node* pn;

	for ( pn = pnFirst; pn; pn = pn->pnNext )
		if ( pn->data == data )
			return true;

	return false;
}
template<class T>
void ValList<T>::Clear()
{
	while( pnFirst )
	{
		Node* pnNext = pnFirst->pnNext;
		delete pnFirst;
		pnFirst = pnNext;
	}

	pnLast = 0;
}

template<class T>
ValList<T>& ValList<T>::operator=(const ValList<T>& rhs)
{
	Clear();

	Append (rhs);

	return *this;
}

#ifdef __FLAT__
template<class T>
BOOL operator==( const typename ValList<T>::ConstIterator& lhs,
				 const typename ValList<T>::Iterator& rhs )
{
	return rhs == lhs;
}

template<typename T>
ValList<T>::Node::Node( const T& data_, Node* pnPrev_, Node* pnNext_) :
data( data_),
pnPrev( pnPrev_ ),
pnNext( pnNext_ )
{
	if( pnPrev )
		pnPrev->pnNext = this;

	if( pnNext )
		pnNext->pnPrev = this;
}

template<typename T>
ValList<T>::Node::~Node()
{
	if( pnPrev )
		pnPrev->pnNext = pnNext;

	if( pnNext )
		pnNext->pnPrev = pnPrev;
}

template<typename T>
T& ValList<T>::ConstIterator::Data()
{
	if( !pnBegin )
		CRASH;

	return pnBegin->data;
}

template<typename T>
ValList<T>::ConstIterator::ConstIterator( Node* pnBegin_ ) :
pnBegin( pnBegin_ )
{}

template<typename T>
typename ValList<T>::ConstIterator& ValList<T>::ConstIterator::operator=( const ConstIterator& rhs )
{
	pnBegin = rhs.pnBegin;
	return *this;
}

template<typename T>
BOOL ValList<T>::ConstIterator::operator==(const ConstIterator& rhs) const
{ return pnBegin == rhs.pnBegin; }

template<typename T>
BOOL ValList<T>::ConstIterator::operator!=(const ConstIterator& rhs) const
{ return !(*this == rhs); }

template<typename T>
const T& ValList<T>::ConstIterator::operator*() const
{
	return const_cast<ConstIterator*>(this)->Data();
}

template<typename T>
typename ValList<T>::ConstIterator& ValList<T>::ConstIterator::operator++()
{
	if( pnBegin )
		pnBegin = pnBegin->pnNext;

	return *this;
}

template<typename T>
typename ValList<T>::ConstIterator& ValList<T>::ConstIterator::operator--()
{
	if( pnBegin )
		pnBegin = pnBegin->pnPrev;

	return *this;
}

template<typename T>
typename ValList<T>::Iterator& ValList<T>::Iterator::operator=( const Iterator& rhs )
{
	base_t::operator=( rhs );
	return *this;
}

template<typename T>
ValList<T>::Iterator::Iterator( Node* pnBegin_ ) :
base_t( pnBegin_ )
{}

template<typename T>
BOOL ValList<T>::Iterator::operator==(const Iterator& rhs) const
{ return base_t::operator==( rhs ); }

template<typename T>
T& ValList<T>::Iterator::operator*()
{
	return Data();
}

template<typename T>
typename ValList<T>::Iterator& ValList<T>::Iterator::operator++()
{
	base_t::operator++();
	return *this;
}

template<typename T>
typename ValList<T>::Iterator& ValList<T>::Iterator::operator--()
{
	base_t::operator--();
	return *this;
}

template<typename T>
ValList<T>::ValList() : pnFirst(0), pnLast(0)
{}

template<typename T>
ValList<T>::ValList(const ValList<T>& rhs) : pnFirst(0), pnLast(0)
{
	*this = rhs;
}

template<typename T>
ValList<T>::~ValList()
{
	Clear();
}

template<typename T>
ValList<T>& ValList<T>::operator<<( const T& data )
{
	Add( data );
	return *this;
}

template<typename T>
BOOL ValList<T>::Contains( const T& data ) const
{
	return Find( data ) != End();
}
#endif __FLAT__

template<typename T>
class Stack : ValList<T>
{
public:
	void Push( const T& data )	{ Add( data ); }
	void Pop()			{ assert( RBegin() != REnd() ); Erase( RBegin() ); }
	const T* Top() const		{ return RBegin() != REnd() ? &*RBegin() : 0;	   }
};

#endif /* _MSC_VER */

//////////////////////////////////////////////////////////////////////////////
//								    DeleteList
// removes and calls delete for each element of a list
//////////////////////////////////////////////////////////////////////////////
template<class Lst>
Lst&
DeleteList(Lst& l)
{
	typename Lst::Element*	pElement;
	while (0 != (pElement = l.PopHead()))
		delete pElement;
	return l;
}


#endif __LISTS_H

