#define MODULE ttfonttest
#include <windows.h>
#include <stdio.h>
#include "ttfonttest.h"
#include "lists.h"

int
FIO_Read(HFILE fd, void * lpData, int len)
{
	return fread(lpData, len, 1, (FILE *)fd);
}

long
FIO_Seek(HFILE fd, long i, int iFrom)
{
	long	lRslt = -1;

	switch(iFrom)
	{
	case	0:
		lRslt = fseek((FILE *)fd, i, SEEK_SET);
		break;
	case	1:
		lRslt = fseek((FILE *)fd, i, SEEK_CUR);
		break;
	case	2:
		lRslt = fseek((FILE *)fd, i, SEEK_END);
		break;
	default:
		break;
	}

	return lRslt;
}

long
FIO_Tell(HFILE fd)
{
	return ftell((FILE *)fd);
}

long
FIO_GetFileSize(HFILE fd)
{
	FILE *fp = (FILE *)fd;
	long pos = ftell(fp);
	long rslt;

	fseek(fp, 0, SEEK_END);
	rslt = ftell(fp);
	fseek(fp, pos, SEEK_SET);

	return rslt;
}

unsigned int
ReadUINT32(HFILE fd)
{
	char	c1, c2, c3, c4;
	unsigned int rslt;

	FIO_Read(fd, (void *)&c1, 1);
	FIO_Read(fd, (void *)&c2, 1);
	FIO_Read(fd, (void *)&c3, 1);
	FIO_Read(fd, (void *)&c4, 1);

	rslt =	((c1 << 24) & 0xff000000) |
		((c2 << 16) & 0x00ff0000) |
		((c3 << 8 ) & 0x0000ff00) |
		( c4	    & 0x000000ff);
	return rslt;
}

unsigned short
ReadUINT16(HFILE fd)
{
	char	c1;
	char	c2;
	unsigned short	rslt;

	FIO_Read(fd, (void *)&c1, 1);
	FIO_Read(fd, (void *)&c2, 1);

	rslt = ((c1 << 8) & 0x0FF00) |
		(c2 & 0x0FF);
	return rslt;
}

int
ReadINT32(HFILE fd)
{
	char	c1, c2, c3, c4;
	int rslt;

	FIO_Read(fd, (void *)&c1, 1);
	FIO_Read(fd, (void *)&c2, 1);
	FIO_Read(fd, (void *)&c3, 1);
	FIO_Read(fd, (void *)&c4, 1);

	rslt =	((c1 << 24) & 0xff000000) |
		((c2 << 16) & 0x00ff0000) |
		((c3 << 8 ) & 0x0000ff00) |
		( c4	    & 0x000000ff);
	return rslt;
}

short
ReadINT16(HFILE fd)
{
	char	c1;
	char	c2;
	short	rslt;

	FIO_Read(fd, (void *)&c1, 1);
	FIO_Read(fd, (void *)&c2, 1);

	rslt = ((c1 << 8) & 0x0FF00) |
		(c2 & 0x0FF);

	return rslt;
}

char
ReadChar(HFILE fd)
{
	char	c;
	FIO_Read(fd, (void *)&c, 1);
	return c;
}

unsigned char
ReadUChar(HFILE fd)
{
	unsigned char c;

	FIO_Read(fd, (void *)&c, 1);

	return c;
}

unsigned char
ReadUINT8(HFILE fd)
{
	unsigned char c;

	FIO_Read(fd, (void *)&c, 1);

	return c;
}

void
ReadString(HFILE fd, char * cp, int len)
{
	FIO_Read(fd, (void *)cp, len);
}

unsigned int
PreReadUINT32(HFILE fd)
{
	unsigned int rslt = ReadUINT32(fd);

	FIO_Seek(fd, -4, 1);

	return rslt;
}

unsigned char
PreReadUChar(HFILE fd)
{
	unsigned char	rslt = ReadUChar(fd);

	FIO_Seek(fd, -1, 1);

	return rslt;
}

unsigned short
PreReadUINT16(HFILE fd)
{
	unsigned short rslt = ReadUINT16(fd);

	FIO_Seek(fd, -2, 1);

	return rslt;
}

void
PreReadString(HFILE fd, char * cp, int len)
{
	long	posi = 0 - len;

	while (len-- > 0)
	{
		FIO_Read(fd, (void *)cp++, 1);
	}

	FIO_Seek(fd, posi, 1);
}

void
Fixed::ReadData(HFILE fd)
{
	high = ReadUINT16(fd);
	low = ReadUINT16(fd);
}

void
LongDateTime::ReadData(HFILE fd)
{
	high = ReadUINT32(fd);
	low = ReadUINT32(fd);
}

TableDirectory::TableDirectory(const TableDirectory &rhs)
	:checkSum(rhs.checkSum),
	 offset(rhs.offset),
	 length(rhs.length)
{
	int	i;

	for (i = 0; i < 4; i++)
		tag[i] = rhs.tag[i];
}

const TableDirectory &
TableDirectory::operator=(const TableDirectory & rhs)
{
	memcpy(tag, rhs.tag, 4);
	checkSum = rhs.checkSum;
	offset = rhs.offset;
	length = rhs.length;

	return *this;
}

void
TableDirectory::ReadTable(HFILE fd)
{
	ReadString(fd, tag, 4);
	checkSum = ReadUINT32(fd);
	offset = ReadUINT32(fd);
	length = ReadUINT32(fd);
}

int
TableDirectory::ReadData(HFILE )
{
	return 1;
}

void
TTFontFile::Reset()
{
	TableDirectory	*lpTD;

	while ((lpTD = lTableDirectories.PopHead()))
		delete lpTD;

	lpCmap = 0;
	lpOS = 0;

	offsetSubtable.Reset();
}

BOOL
TTFontFile::ReadIn(HFILE fd)
{
	Reset();
	BOOL	bRslt = FALSE;
	long	iTotalLen = FIO_GetFileSize(fd);

	if (offsetSubtable.ReadIn(fd))
	{
		char	tag[5];
		UINT32	i;
		Walker<TableDirectory>	tables;
		TableDirectory	*lpTD;
		TableDirectory	tTD;

		bRslt = TRUE;

		tag[4] = '\0';

		for (i = 0; i < offsetSubtable.GetNumTables(); i++)
		{
			PreReadString(fd, tag, 4);

			// Identify Table Entry here
			if (!strcmp(tag, "cmap"))
				lpTD = lpCmap = new CMAP_Directory();
			else if (!strcmp(tag, "OS/2"))
				lpTD = lpOS = new OS2_Directory();
			else
			{
				tTD.ReadTable(fd);
				continue;
			}

			lpTD->ReadTable(fd);
			lTableDirectories.AddElement(lpTD);

			// for Korean fonts using TTC format
			// window doesn't return valid font data
			if (lpTD->GetOffset() > iTotalLen)
				return FALSE;
		}

		for (tables = lTableDirectories; tables; tables++)
		{
			lpTD = tables;

			FIO_Seek(fd, lpTD->GetOffset(), 0);

			if (!lpTD->ReadData(fd))
			{
				bRslt = FALSE;
				break;
			}
		}
	}

	return bRslt;
}

void
TTCFontFile::Reset()
{
	TTFontFile	*pFontFile;

	while((pFontFile = lSubFontFiles.PopHead()))
		delete pFontFile;

	TTFontFile::Reset();
}

BOOL
TTCFontFile::ReadIn(HFILE fd)
{
	if (!TTFontFile::ReadIn(fd))
	{
		FIO_Seek(fd, 0, 0);

		if (!ttc.ReadIn(fd))
			return FALSE;

		int	iCount = ttc.GetDirectoryCount();
		int	i;
		UINT32	iOffset;
		TTFontFile *pFontFile;

		for(i = 0; i < iCount; i++)
		{
			iOffset = ttc.GetTableOffset(i);
			FIO_Seek(fd, iOffset, 0);
			if (i == 0)
			{
				TTFontFile::ReadIn(fd);
			}
			else
			{
				pFontFile = new TTFontFile();
				pFontFile->ReadIn(fd);
				lSubFontFiles.AddElement(pFontFile);
			}
		}
	}
	return TRUE;
}

TTC_Table::~TTC_Table()
{
	Reset();
}

void
TTC_Table::Reset()
{
	iDirectoryCount = 0;
	if (pDirectoryTable)
		delete [] pDirectoryTable;
	pDirectoryTable = 0;
}

int
TTC_Table::ReadIn(HFILE fd)
{
	Reset();

	ReadString(fd, achTag, 4);
	achTag[4] = '\0';

	if (strncmp(achTag, "ttcf", 4))
		return 0;

	iVersion = ReadUINT32(fd);
	iDirectoryCount = ReadUINT32(fd);
	if (iVersion == 0x010000)
	{
		int	i;

		pDirectoryTable = new UINT32[iDirectoryCount + 1];
		for (i = 0; i < iDirectoryCount; i++)
		{
			pDirectoryTable[i] = ReadUINT32(fd);
		}

		pDirectoryTable[i] = FIO_GetFileSize(fd);
	}

	return 1;
}

UINT32
TTC_Table::GetTableOffset(int i)
{
	if (pDirectoryTable && i >= 0 && i <= iDirectoryCount)
		return pDirectoryTable[i];
	return 0;
}
void
OffsetSubtable::Reset()
{
	scaler_type = 0x010000;
	numTables = 0;
	searchRange = 0;
	entrySelector = 0;
	rangeShift = 0;
}

int
OffsetSubtable::ReadIn(HFILE fd)
{
	scaler_type = ReadUINT32(fd);
	numTables = ReadUINT16(fd);
	searchRange = ReadUINT16(fd);
	entrySelector = ReadUINT16(fd);
	rangeShift = ReadUINT16(fd);

	return scaler_type == 0x010000;
}

void
OffsetSubtable::SetScalerType(UINT32 i)
{
	scaler_type = i;
}

void
OffsetSubtable::SetNumTables(UINT32 i)
{
	unsigned int mask = 0x80000000;

	numTables = i;

	i = 31;

	while (mask && !(numTables & mask))
	{
		i--;
		mask >>= 1;
	}

	searchRange = mask * 16;
	entrySelector = i;
	rangeShift = numTables * 16 - searchRange;
}


int
TTFontFile::GetFirstChar()
{
	return lpOS ? lpOS->fsFirstCharIndex : -1;
}

int
TTFontFile::GetLastChar()
{
	return lpOS ? lpOS->fsLastCharIndex : -1;
}

int
TTFontFile::GetCodePageRange1()
{
	return lpOS ? lpOS->ulCodePageRange1 : -1;
}

int
TTFontFile::GetCodePageRange2()
{
	return lpOS ? lpOS->ulCodePageRange2 : -1;
}

BOOL
TTFontFile::HasSymbolCmap()
{
	return lpCmap ? lpCmap->HasSymbolCmap() : FALSE;
}

void
OS2_Directory::PANOSE::ReadData(HFILE fd)
{
	bFamilyType = ReadUINT8(fd);
	bSerifStyle = ReadUINT8(fd);
	bWeight = ReadUINT8(fd);
	bProportion = ReadUINT8(fd);
	bContrast = ReadUINT8(fd);
	bStrokeVariation = ReadUINT8(fd);
	bArmStyle = ReadUINT8(fd);
	bLetterform = ReadUINT8(fd);
	bMidline = ReadUINT8(fd);
	bXHeight = ReadUINT8(fd);
}

int
OS2_Directory::ReadData(HFILE  fd)
{
	int	i;
	long	cur_pos = FIO_Tell(fd);

	version = ReadUINT16(fd);
	xAveCharWidth = ReadUINT16(fd);
	usWeightClass = ReadUINT16(fd);
	usWidthClass = ReadUINT16(fd);
	fsType1 = ReadUINT8(fd);
	fsType2 = ReadUINT8(fd);
	ySubscriptXSize = ReadUINT16(fd);
	ySubscriptYSize = ReadUINT16(fd);
	ySubscriptXOffset = ReadUINT16(fd);
	ySubscriptYOffset = ReadUINT16(fd);
	ySuperscriptXSize = ReadUINT16(fd);
	ySuperscriptYSize = ReadUINT16(fd);
	ySuperscriptXOffset = ReadUINT16(fd);
	ySuperscriptYOffset = ReadUINT16(fd);
	yStrikeoutSize = ReadUINT16(fd);
	yStrikeoutPosition = ReadUINT16(fd);
	sFamilyClass = ReadUINT8(fd);
	sFamilySubClass = ReadUINT8(fd);
	panose.ReadData(fd);
	for (i = 0; i < 4; i++)
		ulCharRange[i] = ReadUINT32(fd);
	for (i = 0; i < 4; i++)
		achVendID[i] = ReadUINT8(fd);
	fsSelection = ReadUINT16(fd);
	fsFirstCharIndex = ReadUINT16(fd);
	fsLastCharIndex = ReadUINT16(fd);
	sTypoAscender = ReadUINT16(fd);
	sTypoDescender = ReadUINT16(fd);
	sTypoLineGap = ReadUINT16(fd);
	usWinAscent = ReadUINT16(fd);
	usWinDescent = ReadUINT16(fd);
	ulCodePageRange1 = ReadUINT32(fd);
	ulCodePageRange2 = ReadUINT32(fd);

	if (version == 2)
	{
		xHeight = ReadUINT16(fd);
		CapHeight = ReadUINT16(fd);
		defaultChar = ReadUINT16(fd);
		breakChar = ReadUINT16(fd);
		maxLookups = ReadUINT16(fd);
	}
	else
	{
		version = 0;

		xHeight = 0;
		CapHeight = 0;
		defaultChar = 0;
		breakChar = 0;
		maxLookups = 0;
	}

	FIO_Seek(fd, cur_pos, 0);

	return TableDirectory::ReadData(fd);
}


CMAP_Directory::~CMAP_Directory()
{
	CMAP_Subtable	*lpST;

	while ((lpST = lSubtables.PopHead()))
		delete lpST;
}

BOOL
CMAP_Directory::ReadData(HFILE fd)
{
	long	cur_pos = FIO_Tell(fd);

	BOOL	rslt = TRUE;
	int	i;
	CMAP_Subtable	*lpST;

	iVersion = ReadUINT16(fd);
	numSubtables = ReadUINT16(fd);

	for (i = 0; i < numSubtables; i++)
	{
		lpST = new CMAP_Subtable(this);
		rslt |= lpST->ReadIn(fd);
		lSubtables.AddElement(lpST);
	}

	FIO_Seek(fd, cur_pos, 0);

	return TableDirectory::ReadData(fd) && rslt;
}

BOOL
CMAP_Directory::CMAP_Subtable::ReadIn(HFILE fd)
{
	long	cur_pos;
	long	data_offset;

	platformID = ReadUINT16(fd);
	platformSpecificID = ReadUINT16(fd);
	offset = ReadUINT32(fd);
	cur_pos = FIO_Tell(fd);

	data_offset = offset + lpParent->GetOffset();

	FIO_Seek(fd, data_offset, 0);

	form_id = ReadUINT16(fd);

	FIO_Seek(fd, cur_pos, 0);

	return TRUE;
}

BOOL
CMAP_Directory::HasSymbolCmap()
{
	BOOL	bRslt = FALSE;
	Walker<CMAP_Subtable>	subtables;

	for (subtables = lSubtables; subtables; subtables++)
	{
		if (subtables->IsSymbolCmap())
		{
			bRslt = TRUE;
			break;
		}
	}

	return bRslt;
}


BOOL
CheckFontInfo(HDC hDC, HFONT hFont, FILE * fp, TTFontFile & ttf)
{
	TEXTMETRIC tmA;
	TEXTMETRICW tmW;
	LOGFONT lf;

	GetObject((HGDIOBJ)hFont, sizeof(LOGFONT), &lf);
	if (!GetTextMetrics(hDC, &tmA))
	{
		DWORD iError = GetLastError();
		fprintf(fp, "GetTextMetrics() Error: %x\n", iError);
		return FALSE;
	}

	if (!GetTextMetricsW(hDC, &tmW))
	{
		DWORD iError = GetLastError();
		fprintf(fp, "Error: %x\n", iError);
		return FALSE;
	}

	fprintf(fp, "                     GetTextMetrics   |     GetTextMetricsW     |      TTFontFile  \n");
	fprintf(fp, "FirstChar               %d\t      |         %d\t        |        % 5d.%04x\n", tmA.tmFirstChar, tmW.tmFirstChar, ttf.GetFirstChar(), ttf.GetFirstChar());
	fprintf(fp, "LastChar                %d\t      |         %05d.%04x\t|       % 5d.%04x\n", tmA.tmLastChar, tmW.tmLastChar, tmW.tmLastChar, ttf.GetLastChar(),  ttf.GetLastChar());
	fprintf(fp, "DefaultChar             %d\t      |         %d\t        |       \n",     tmA.tmDefaultChar, tmW.tmDefaultChar);
	fprintf(fp, "BreakChar               %d\t      |         %d\t        |       \n",     tmA.tmBreakChar, tmW.tmBreakChar);
	fprintf(fp, "CharSet/CodeRange       %d\t      |         %d\t        |       %08x.%08x\n", tmA.tmCharSet, tmW.tmCharSet,
												ttf.GetCodePageRange1(), ttf.GetCodePageRange2());

	// Basic rule
	// tmA.tmBreakChar = tmW.tmBreakChar
	if (tmA.tmBreakChar != tmW.tmBreakChar)
		fprintf(fp, "P0: tmA.BreakChar != tmW.BreakChar\n");
	/**
	 * for Symbol CharSet
	 */
	if (tmA.tmCharSet == SYMBOL_CHARSET)
	{
		/* W version */
		//	tmW.tmFirstChar = 0;
		//	tmW.tmLastChar = 0x0F0FF;
		//	tmW.tmBreakChar = pOS2->fsFirstCharIndex & 0x0FF;
		//	tmW.tmDefaultChar = tmW.tmBreakChar - 1
		if (tmW.tmFirstChar != 0)
			fprintf(fp, "P1: tmW.FirstChar != 0 while SYMBOL charset\n");
		if (tmW.tmLastChar != 0x0F0FF)
			fprintf(fp, "P2: tmW.LastChar != 0x0F0FF while SYMBOL charset\n");
		if (tmW.tmBreakChar != (ttf.GetFirstChar() & 0x0FF))
			fprintf(fp, "P3: tmW.BreakChar != (0x0FF & ttf.GetFirstChar())\n");
		if (tmW.tmBreakChar != (tmW.tmDefaultChar + 1))
			fprintf(fp, "P4: tmW.BreakChar != tmW.DefaultChar +1\n");

		/* A version */
		//	tmA.tmBreakChar = tmW.tmBreakChar;
		//	tmA.tmFirstChar = pOS2->fsFirstCharIndex & 0x0FF - 2; (tmA.tmFirstChar = tmW.tmBreakChar - 2)
		//	tmA.tmLastChar = pOS2->fsLastCharIndex > 0x0F0FF ? 255 : pOS2->fsLastCharIndex & 0x0FF;
		if (tmA.tmFirstChar != ((ttf.GetFirstChar() & 0x0FF) - 2))
			fprintf(fp, "P5: tmA.FirstChar != (ttf.GetFirstChar() & 0x0FF) - 2\n");
		if (ttf.GetLastChar() > 0x0F0FF)
		{
			if (tmA.tmLastChar != 0x0FF)
				fprintf(fp, "P6: tmA.LastChar != 0x0FF while ttf.GetLastChar > 0x0F0FF \n");
		}
		else
		{
			if (tmA.tmLastChar != (ttf.GetLastChar() & 0x0FF))
				fprintf(fp, "P7: tmA.LastChar != ttf.GetLastChar & 0x0FF\n");
		}
	}
	else
	{
		/* W version */
		//	tmW.tmBreakChar = pOS2->fsFirstCharIndex;
		//	if (tmW.tmBreakChar < 2)
		//		tmW.tmBreakChar += 2;
		//	tmW.tmFirstChar = pOS2->fsFirstCharIndex;
		//	tmW.tmLastChar = pOS->fsLastCharIndex;
		if (tmW.tmBreakChar != ttf.GetFirstChar())
		{
			if (ttf.GetFirstChar() < 2)
			{
				if (tmW.tmBreakChar != (ttf.GetFirstChar() + 2))
					fprintf(fp, "P8: tmW.FirstChar != ttf.GetFirstChar() + 2 while ttf firstchar < 2\n");
			}
			else
				fprintf(fp, "P9: tmW.FirstChar != ttf.GetFirstChar() while not SYMBOL font\n");
		}
		if (tmW.tmFirstChar != ttf.GetFirstChar())
			fprintf(fp, "P10: tmW.FirstChar != ttf.GetFirstChar()\n");
		if (tmW.tmLastChar != ttf.GetLastChar())
			fprintf(fp, "P11: tmW.LastChar != ttf.GetLastChar()\n");

		/* A version */
		//	tmA.tmBreakChar = tmW.tmBreakChar;
		//	tmA.tmFirstChar = tmA.tmBreakChar - 2;
		//	tmA.tmLastChar = tmW.tmLastChar > 255 ? 255 : tmW.tmLastChar;
		if (tmA.tmFirstChar != (tmA.tmBreakChar - 2))
			fprintf(fp, "P12: tmA.FirstChar != tmA.BreakChar - 2\n");
		if (tmW.tmLastChar > 255)
		{
			if (tmA.tmLastChar != 255)
				fprintf(fp, "P13: tmA.LastChar != 255 while tmW.LastChar > 255\n");
		}
		else
		{
			if (tmA.tmLastChar != tmW.tmLastChar)
				fprintf(fp, "P14: tmA.LastChar != tmW.tmLastChar\n");
		}

		if (tmW.tmLastChar > 255)
		{
			if (tmA.tmLastChar != 255)
				fprintf(fp, "P15: tmA.LastChar != 255 while tmW.LastChar > 255\n");
		}
		else
		{
			if (tmA.tmLastChar != tmW.tmLastChar)
				fprintf(fp, "P16: tmA.LastChar != tmW.LastChar\n");
		}
	}

	//	tmA.tmDefaultChar = tmW.tmDefaultChar > 255 ? tmA.tmBreakChar : tmW.tmDefaultChar;
	if (tmW.tmDefaultChar > 255)
	{
		if (tmA.tmDefaultChar != tmA.tmBreakChar)
			fprintf(fp, "P17: tmA.DefaultChar != 255 while tmW.DefaultChar > 255\n");
	}
	else
	{
		if (tmA.tmDefaultChar != tmW.tmDefaultChar)
			fprintf(fp, "P18: tmA.DefaultChar != tmW.DefaultChar\n");
	}

	//	tmW.tmDefaultChar = tmW.tmBreakChar - 1;
	if (tmW.tmDefaultChar != (tmW.tmBreakChar - 1))
		fprintf(fp, "P19: tmW.DefaultChar != tmW.BreakChar - 1 Def %d Brk %d\n", tmW.tmDefaultChar, tmW.tmBreakChar);
	
	// Extra Rule
	// if TrueTypeFont has Symbol CMAP section (Format 4), it should be treated as Symbol charset regardless pOS2->ulCodePageRange1
	if (tmA.tmCharSet == 2 && !(ttf.GetCodePageRange1() & 0x80000000))
		fprintf(fp, "P20: ttf %s have Symbol CMAP while pOS2->ulCodePageRange1 doesn't have SYMBOL flag\n",
				ttf.HasSymbolCmap() ? "Does" : "Doesn't");
	else if (tmA.tmCharSet != 2 && ttf.HasSymbolCmap())
		fprintf(fp, "P21: ttf doesnt have Symbol CMAP but it's Symbol CharSet\n");

	fprintf(fp, "%s have Symbol CMAP while CharSet == %d\n", ttf.HasSymbolCmap() ? "Does" : "Doesn't", tmA.tmCharSet);
	fprintf(fp, "\n\n\n");

	return TRUE;
}

HDC	hScreenDC;
FILE	*fpLog;

const char * achTmpFile = "cvj_ttf.ttf";

int CALLBACK
MyEnumFontsProc(const LOGFONT* lplf, const TEXTMETRIC *tm, DWORD, LPARAM)
{
	if (tm->tmPitchAndFamily & TMPF_TRUETYPE)
	{
		FILE	*fp = fopen(achTmpFile, "wb+");

		if (fp)
		{
			fprintf(fpLog, "CheckFont: %s\n", lplf->lfFaceName);

			HFONT	hFont = CreateFontIndirect(lplf);
			HFONT	hOldFont;

			if (hFont == NULL)
			{
				DWORD err = GetLastError();
				fprintf(fpLog, "CreateFontIndirect failed error code %d\n", err);
				fclose(fp);
				return 1;
			}

			hOldFont = (HFONT)SelectObject(hScreenDC, hFont);

			int	iLen = GetFontData(hScreenDC, 0x66637474, 0, 0, 0);
			BOOL	bTTC = TRUE;

			if (iLen == GDI_ERROR)
			{
				iLen = GetFontData(hScreenDC, 0, 0, 0, 0);
				bTTC = FALSE;
			}

			if (iLen != GDI_ERROR)
			{
				char	*pch = new char[iLen];

				iLen = GetFontData(hScreenDC, bTTC ? 0x66637474 : 0, 0, pch, iLen);

				fwrite(pch, iLen, 1, fp);

				fclose(fp);

				delete [] pch;

				TTCFontFile	fontFile;

				fp = fopen(achTmpFile, "r+b");

				if (!fontFile.ReadIn((HFILE)fp))
					fprintf(fpLog, "Read TTF file failed\n");

				CheckFontInfo(hScreenDC, hFont, fpLog, fontFile);

				fclose(fp);

				unlink(achTmpFile);
			}
			else
			{
				DWORD err = GetLastError();
				fprintf(fpLog, "cannot get font data error code %d\n", err);

				fclose(fp);
			}

			SelectObject(hScreenDC, hOldFont);
			DeleteObject(hFont);
		}
	}
	return 1;
}

#ifdef linux
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
		   LPSTR lpCmdLine, int nCmdShow)
#else
int	main()
#endif
{
#ifdef linux
	FILE	*fp = fopen("tm_linux.txt", "w+");
#else
	FILE	*fp = fopen("tm_window.txt", "w+");
#endif
	HDC	hDC = GetDC(NULL);

	if (hDC)
	{
		hScreenDC = hDC;

		if (fp)
		{
			fpLog = fp;

			fprintf(fp, "Check points:\n");
			fprintf(fp, "P0: tmA.BreakChar != tmW.BreakChar\n");
			fprintf(fp, "P1: tmW.FirstChar != 0 while SYMBOL charset\n");
			fprintf(fp, "P2: tmW.LastChar != 0x0F0FF while SYMBOL charset\n");
			fprintf(fp, "P3: tmW.BreakChar != (0x0FF & ttf.GetFirstChar())\n");
			fprintf(fp, "P4: tmW.BreakChar != tmW.DefaultChar +1\n");
			fprintf(fp, "P5: tmA.FirstChar != (ttf.GetFirstChar() & 0x0FF) - 2\n");
			fprintf(fp, "P6: tmA.LastChar != 0x0FF while ttf.GetLastChar > 0x0F0FF \n");
			fprintf(fp, "P7: tmA.LastChar != ttf.GetLastChar & 0x0FF\n");
			fprintf(fp, "P8: tmW.FirstChar != ttf.GetFirstChar() + 2 while ttf firstchar < 2\n");
			fprintf(fp, "P9: tmW.FirstChar != ttf.GetFirstChar() while not SYMBOL font\n");
			fprintf(fp, "P10: tmW.FirstChar != ttf.GetFirstChar()\n");
			fprintf(fp, "P11: tmW.LastChar != ttf.GetLastChar()\n");
			fprintf(fp, "P12: tmA.FirstChar != tmA.BreakChar - 2\n");
			fprintf(fp, "P13: tmA.LastChar != 255 while tmW.LastChar > 255\n");
			fprintf(fp, "P14: tmA.LastChar != tmW.tmLastChar\n");
			fprintf(fp, "P15: tmA.LastChar != 255 while tmW.LastChar > 255\n");
			fprintf(fp, "P16: tmA.LastChar != tmW.LastChar\n");
			fprintf(fp, "P17: tmA.DefaultChar != 255 while tmW.DefaultChar > 255\n");
			fprintf(fp, "P18: tmA.DefaultChar != tmW.DefaultChar\n");
			fprintf(fp, "P19: tmW.DefaultChar != tmW.BreakChar - 1 Def ? Brk ?\n");
			fprintf(fp, "P20: if ttf have Symbol CMAP while pOS2->ulCodePageRange1 doesn't have SYMBOL flag\n");
			fprintf(fp, "P21: ttf doesnt have Symbol CMAP but it's Symbol CharSet\n");
			fprintf(fp, "\n\n");

			EnumFonts(hScreenDC, 0, MyEnumFontsProc, 0);

			fclose(fp);
		}

		DeleteDC(hScreenDC);
	}

	return 0;
}

