#ifndef __TTFONTFILE_H
#define __TTFONTFILE_H

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lists.h"

#ifndef UINT16
typedef unsigned short	UINT16;
#endif

#ifndef UINT32
typedef unsigned int	UINT32;
#endif

#ifndef UINT8
typedef unsigned char	UINT8;
#endif

typedef short		FWord;
typedef unsigned short	uFWord;

unsigned int	ReadUINT32(HFILE fd);
unsigned short	ReadUINT16(HFILE fd);
char		ReadChar(HFILE fd);
unsigned char	ReadUChar(HFILE fd);
unsigned char	ReadUINT8(HFILE fd);
void		ReadString(HFILE fd, char *, int);

int		ReadINT32(HFILE);
short		ReadINT16(HFILE);


void		PreReadString(HFILE fd, char *, int);
unsigned int	PreReadUINT32(HFILE fd);
unsigned short	PreReadUINT16(HFILE fd);
unsigned char	PreReadUChar(HFILE fd);

class	Fixed
{
public:
	UINT16	high;
	UINT16	low;

	Fixed()
	:high(0),
	 low(0)
	{}

	Fixed(const Fixed & rhs)
	:high(rhs.high),
	 low(rhs.low)
	{}

	void	ReadData(HFILE fp);
};

class	LongDateTime
{
public:
	UINT32	high;
	UINT32	low;

	LongDateTime()
	:high(0),
	 low(0)
	{}

	LongDateTime(const LongDateTime & rhs)
	:high(rhs.high),
	 low(rhs.low)
	{}

	void	ReadData(HFILE fp);
};

/********************************************************************************
 *
 *	TrueType Collection (TTC) is a means of delivering multiple TrueType
 *	fonts in a single file structure. TrueType Collections are most useful
 *	when the fonts to be delivered together share many glyphs in common.
 *
 *	TAG	TTCTage		TrueTypeCollection ID String: "ttcf"
 *	FIXED32 version		Version of TTC Table (initially 0x00010000)
 *	ULONG	DirectoryCount	Number of Table Directories in TTC
 *	ULONG	TableDirectory	Array of offsets to TableDirectory from file begin
		[DirectoryCount]
 *******************************************************************************/

class TTC_Table
{
	char	achTag[5];
	UINT32	iVersion;
	UINT32	iDirectoryCount;
	UINT32	*pDirectoryTable;
public:
	TTC_Table()
	:iVersion(0x010000),
	 iDirectoryCount(0),
	 pDirectoryTable(0)
	{}

	~TTC_Table();

	void	Reset();

	int	ReadIn(HFILE);

	UINT32	GetDirectoryCount()	{ return iDirectoryCount;}
	UINT32	GetTableOffset(int);
};



/********************************************************************************/
/*	First talble describe num of subtable info in TrueTypeFont file		*/
/*										*/
/*	uint32	scaler_type	A tag to indicate the OFA scaler to be	used to */
/*				rasterize this font; see the note on the scaler */
/*				type below for more information.		*/
/*	uint16	numTables	number of tables				*/
/*	uint16	searchRange	(maximum power of 2 <= numTables)*16		*/
/*	uint16	entrySelector	log2(maximum power of 2 <= numTables)		*/
/*	uint16	rangeShift	numTables*16-searchRange			*/
/*										*/
/*										*/
/*	e.g numTables:		25						*/
/*	    maximum power of 2: 16						*/
/*	    searchRange:	16 * 16 = 256					*/
/*	    entrySelector:	4						*/
/*	    rangeShift:		25 * 16 - searchRange = 144			*/
/*										*/
/*										*/
/*	    numTables:		13						*/
/*	    maximum power of 2: 8						*/
/*	    searchRange:	16 * 8 = 128					*/
/*	    entrySelector:	3						*/
/*	    rangeShift:		13 * 16 - 128 = 80				*/
/*										*/
/********************************************************************************/
class	OffsetSubtable
{
	UINT32	scaler_type;
	UINT16	numTables;
	UINT16	searchRange;
	UINT16	entrySelector;
	UINT16	rangeShift;
public:
	OffsetSubtable()
	:scaler_type(0x010000),
	 numTables(0),
	 searchRange(0),
	 entrySelector(0),
	 rangeShift(0)
	{}

	void	Reset();

	int	ReadIn(HFILE);

	void	SetScalerType(UINT32);
	void	SetNumTables(UINT32);

	UINT32	GetNumTables()	{ return numTables;}
};

/********************************************************************************/
/*	The table directory follows the offset subtable. Entries in the table	*/
/*	directory must be sorted in ascending order by tag. Each table in the	*/
/*	font file must have its own table directory entry. Table 5 documents	*/
/*	the structure of the table directory.					*/
/*										*/
/*	uint32	tag		4-byte identifier				*/
/*	uint32	checkSum	checksum for this table				*/
/*	uint32	offset		offset from beginning of sfnt			*/
/*	uint32	length		length of this table in byte (actual length not */
/*				padded length)					*/
/********************************************************************************/
class TableDirectory
{
protected:
	char		tag[4];
	unsigned int	checkSum;
	unsigned int	offset;
	unsigned int	length;

public:

	TableDirectory()
	:checkSum(0),
	 offset(0),
	 length(0)
	{
		memset(tag, 0, 4);
	}

	TableDirectory(const TableDirectory &rhs);

	virtual ~TableDirectory()
	{
		Reset();
	}

	const TableDirectory & operator=(const TableDirectory &);

	virtual void	Reset() {}

	// it always be two stages process:
	//	ReadIn:		read table info
	//	ReadData:	read real chunk of data
	virtual void	ReadTable(HFILE);
	virtual int	ReadData(HFILE);

	virtual const char *	GetTagName()	{ return tag;}

	virtual int		GetOffset()	{ return offset;}
};


class	CMAP_Directory;
class	OS2_Directory;

class	TTFontFile
{
protected:
	CMAP_Directory	*lpCmap;
	OS2_Directory	*lpOS;

public:
	OffsetSubtable	offsetSubtable;
	List<TableDirectory>	lTableDirectories;

	TTFontFile()
	{
		Reset();
	}

	virtual ~TTFontFile()
	{
		Reset();
	}

	virtual void	Reset();

	virtual BOOL	ReadIn(HFILE);

	int		GetFirstChar();
	int		GetLastChar();
	int		GetCodePageRange1();
	int		GetCodePageRange2();
	BOOL		HasSymbolCmap();
	void		MakeSymbolCmap();
};

class	TTCFontFile: public TTFontFile
{
	TTC_Table		ttc;
	List<TTFontFile>	lSubFontFiles;
public:
	~TTCFontFile()
	{
		Reset();
	}

	void	Reset();

	BOOL	ReadIn(HFILE);

	BOOL		IsTTCFormat() { return lSubFontFiles.GetHead() ? TRUE:FALSE;}
};

/********************************************************************************/
/*	The format of this table has changed over time. The original TrueType	*/
/*	specification had this table at 68 bytes long. The first Microsoft	*/
/*	version had it at 78 bytes long, and the current Microsoft version	*/
/*	is 86 bytes long. To determine which kind of table your software is	*/
/*	dealing with, look at the table size using these numbers as a guide.	*/
/*										*/
/*	This class definition is based on 68 byte version.			*/
/********************************************************************************/

//Class for TTFontFile 'OS/2'
//	UINT16	version				(set to zero)
//	UINT16	xAveCharWidth			average weighted advance width of lower case letters and space
//	UINT16	usWeightClass			visual weight (degree of blackness or thickness) of stroke in glyphs
//	UINT16	usWidthClass			relative change from the normal aspect ratio (width to height ratio) as specified by a font designer for
//						the glyphs in the font
//	UINT16	fsType				characteristics and properties of this font (set undefined bits to zero)
//	UINT16	ySubscriptXSize			recommended horizontal size in pixels for subscripts
//	UINT16	ySubscriptYSize			recommended vertical size in pixels for subscripts
//	UINT16	ySubscriptXOffset		recommended horizontal offset for subscripts
//	UINT16	ySubscriptYOffset		recommended vertical offset form the baseline for subscripts
//	UINT16	ySuperscriptXSize		recommended horizontal size in pixels for superscripts
//	UINT16	ySuperscriptYSize		recommended vertical size in pixels for superscripts
//	UINT16	ySuperscriptXOffset		recommended horizontal offset for superscripts
//	UINT16	ySuperscriptYOffset		recommended vertical offset from the baseline for superscripts
//	UINT16	yStrikeoutSize			width of the strikeout stroke
//	UINT16	yStrikeoutPosition		position of the strikeout stroke relative to the baseline
//	UINT16	sFamilyClass			classification of font-family design.
//	PANOSE	panose				10 byte series of number used to describe the visual characteristics of a given typeface
//	UINT32	ulCharRange[4]			Field is split into two bit fields of 96 and 36 bits each. The low 96 bits are used to specify the Unicode
//						blocks encompassed by the font file. The high 32 bits are used to specify the character or script sets
//						covered by the font file. Bit assignments are pending. Set to 0
//	UINT8	achVendID[4]			four character identifier for the font vendor
//	UINT16	fsSelection			2-byte bit field containing information concerning the nature of the font patterns
//	UINT16	fsFirstCharIndex		The minimum Unicode index in this font.
//	UINT16	fsLastCharIndex			The maximum Unicode index in this font.
//	UINT16	sTypoAscender			
//	UINT16	sTypoDescender
//	UINT16	sTypoLineGap
//	UINT16	usWinAscent
//	UINT16	usWinDescent
//	UINT32	ulCodePageRange1
//	UINT32	ulCodePageRange2
//	xHeight
//	CapHeight
//	defaultChar
//	breakChar
//	maxLookups
//	
class	OS2_Directory: public TableDirectory
{
	class	PANOSE
	{
	public:
		UINT8	bFamilyType;
		UINT8	bSerifStyle;
		UINT8	bWeight;
		UINT8	bProportion;
		UINT8	bContrast;
		UINT8	bStrokeVariation;
		UINT8	bArmStyle;
		UINT8	bLetterform;
		UINT8	bMidline;
		UINT8	bXHeight;

		void	ReadData(HFILE fd);
		void	Print();
	};
public:
	UINT16	version;
	UINT16	xAveCharWidth;
	UINT16	usWeightClass;
	UINT16	usWidthClass;
	UINT8	fsType1;		// TODO bits need swap
	UINT8	fsType2;
	UINT16	ySubscriptXSize;
	UINT16	ySubscriptYSize;
	UINT16	ySubscriptXOffset;
	UINT16	ySubscriptYOffset;
	UINT16	ySuperscriptXSize;
	UINT16	ySuperscriptYSize;
	UINT16	ySuperscriptXOffset;
	UINT16	ySuperscriptYOffset;
	UINT16	yStrikeoutSize;
	UINT16	yStrikeoutPosition;
	UINT8	sFamilyClass;
	UINT8	sFamilySubClass;
	PANOSE	panose;
	UINT32	ulCharRange[4];
	UINT8	achVendID[4];
	UINT16	fsSelection;
	UINT16	fsFirstCharIndex;
	UINT16	fsLastCharIndex;
	UINT16	sTypoAscender;
	UINT16	sTypoDescender;
	UINT16	sTypoLineGap;
	UINT16	usWinAscent;
	UINT16	usWinDescent;
	UINT32	ulCodePageRange1;
	UINT32	ulCodePageRange2;

	UINT16	xHeight;
	UINT16	CapHeight;
	UINT16	defaultChar;
	UINT16	breakChar;
	UINT16	maxLookups;

	int	ReadData(HFILE fd);

	const char *	GetTagName()	{ return "OS/2"; }
};

// Class for TTFontFile 'cmap'
//	UINT16	version (set to zero)
//	UINT16	number of subtables
class	CMAP_Directory: public TableDirectory
{
	// class for TrueType font file 'cmap' subtable entry
	//	UINT16	platformID
	//	UINT16	platformSpecificID
	//	UINT32	offset of mapping table
	class	CMAP_Subtable
	{
	public:
		UINT16	platformID;
		UINT16	platformSpecificID;
		UINT32	offset;
		UINT16	form_id;

		CMAP_Directory		*lpParent;

		CMAP_Subtable(CMAP_Directory *_lpParent)
		:platformID(1),
		 platformSpecificID(0),
		 offset(0),
		 lpParent(_lpParent)
		{}

		~CMAP_Subtable()
		{
			Reset();
		}

		void	Reset()
		{
			platformID = 0;
			platformSpecificID = 0;
			offset = 0;
		}

		BOOL		ReadIn(HFILE);
		BOOL		IsSymbolCmap() { return (platformID == 3) && (platformSpecificID == 0);}
	};

public:
	UINT16		iVersion;
	UINT16		numSubtables;

	List<CMAP_Subtable>	lSubtables;

	CMAP_Directory()
	:iVersion(0),
	 numSubtables(0)
	{}

	~CMAP_Directory();

	const char * GetTagName()	{ return "cmap";}

	BOOL		ReadData(HFILE);

	BOOL		HasSymbolCmap();
};




#endif
